@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1995, 1996, 1997, 1998, 1999 Kungliga Tekniska Hgskolan
+ * Copyright (c) 1995 - 2000 Kungliga Tekniska Hgskolan
  * (Royal Institute of Technology, Stockholm, Sweden).
  * All rights reserved.
  * 
@@ -14,12 +14,7 @@
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
  * 
- * 3. All advertising materials mentioning features or use of this software
- *    must display the following acknowledgement:
- *      This product includes software developed by the Kungliga Tekniska
- *      Hgskolan and its contributors.
- * 
- * 4. Neither the name of the Institute nor the names of its contributors
+ * 3. Neither the name of the Institute nor the names of its contributors
  *    may be used to endorse or promote products derived from this software
  *    without specific prior written permission.
  * 
@@ -38,7 +33,7 @@
 
 #include "kx.h"
 
-RCSID("$Id: common.c,v 1.46 1999/04/07 18:39:21 assar Exp $");
+RCSID("$Id: common.c,v 1.61 2001/01/17 10:09:16 joda Exp $");
 
 char x_socket[MaxPathLen];
 
@@ -50,72 +45,6 @@ int xauthfile_size = sizeof(xauthfile);
 u_char cookie[16];
 size_t cookie_len = sizeof(cookie);
 
-static int
-do_enccopy (int fd1, int fd2, int mode, des_cblock *iv,
-	    des_key_schedule schedule, int *num)
-{
-     int ret;
-     u_char buf[BUFSIZ];
-
-     ret = read (fd1, buf, sizeof(buf));
-     if (ret == 0)
-	  return 0;
-     if (ret < 0) {
-	 warn ("read");
-	 return ret;
-     }
-#ifndef NOENCRYPTION
-     des_cfb64_encrypt (buf, buf, ret, schedule, iv,
-			num, mode);
-#endif
-     ret = krb_net_write (fd2, buf, ret);
-     if (ret < 0) {
-	 warn ("write");
-	 return ret;
-     }
-     return 1;
-}
-
-/*
- * Copy data from `fd1' to `fd2', encrypting it.  Data in the other
- * direction is of course, decrypted.
- */
-
-int
-copy_encrypted (int fd1, int fd2, des_cblock *iv,
-		des_key_schedule schedule)
-{
-     des_cblock iv1, iv2;
-     int num1 = 0, num2 = 0;
-
-     memcpy (iv1, *iv, sizeof(iv1));
-     memcpy (iv2, *iv, sizeof(iv2));
-     for (;;) {
-	  fd_set fdset;
-	  int ret;
-
-	  FD_ZERO(&fdset);
-	  FD_SET(fd1, &fdset);
-	  FD_SET(fd2, &fdset);
-
-	  ret = select (max(fd1, fd2)+1, &fdset, NULL, NULL, NULL);
-	  if (ret < 0 && errno != EINTR) {
-	      warn ("select");
-	      return 1;
-	  }
-	  if (FD_ISSET(fd1, &fdset)) {
-	       ret = do_enccopy (fd1, fd2, DES_ENCRYPT, &iv1, schedule, &num1);
-	       if (ret <= 0)
-		    return ret;
-	  }
-	  if (FD_ISSET(fd2, &fdset)) {
-	       ret = do_enccopy (fd2, fd1, DES_DECRYPT, &iv2, schedule, &num2);
-	       if (ret <= 0)
-		    return ret;
-	  }
-     }
-}
-
 #ifndef X_UNIX_PATH
 #define X_UNIX_PATH "/tmp/.X11-unix/X"
 #endif
@@ -124,17 +53,13 @@ copy_encrypted (int fd1, int fd2, des_cblock *iv,
 #define X_PIPE_PATH "/tmp/.X11-pipe/X"
 #endif
 
-#ifndef INADDR_LOOPBACK
-#define INADDR_LOOPBACK 0x7f000001
-#endif
-
 /*
- * Allocate a unix domain socket.
+ * Allocate a unix domain socket in `s' for display `dpy' and with
+ * filename `pattern'
  *
  * 0 if all is OK
  * -1 if bind failed badly
- * 1 if dpy is already used
- */
+ * 1 if dpy is already used */
 
 static int
 try_socket (struct x_socket *s, int dpy, const char *pattern)
@@ -189,11 +114,12 @@ try_pipe (struct x_socket *s, int dpy, const char *pattern)
     
     snprintf (path, sizeof(path), pattern, dpy);
     fd = open (path, O_WRONLY | O_CREAT | O_EXCL, 0600);
-    if (fd < 0)
+    if (fd < 0) {
 	if (errno == EEXIST)
 	    return 1;
 	else
 	    return -1;
+    }
 
     close (fd);
 
@@ -222,6 +148,14 @@ try_pipe (struct x_socket *s, int dpy, const char *pattern)
 }
 #endif /* MAY_HAVE_X11_PIPES */
 
+/*
+ * Try to create a TCP socket in `s' corresponding to display `dpy'.
+ *
+ * 0 if all is OK
+ * -1 if bind failed badly
+ * 1 if dpy is already used
+ */
+
 static int
 try_tcp (struct x_socket *s, int dpy)
 {
@@ -259,8 +193,7 @@ try_tcp (struct x_socket *s, int dpy)
 }
 
 /*
- * Allocate and listen on a number of local X server sockets, pipes,
- * and a TCP socket.  Return the display number.
+ * The potential places to create unix sockets.
  */
 
 static char *x_sockets[] = {
@@ -270,6 +203,10 @@ X_UNIX_PATH "%u",
 NULL
 };
 
+/*
+ * Dito for stream pipes.
+ */
+
 #ifdef MAY_HAVE_X11_PIPES
 static char *x_pipes[] = {
 X_PIPE_PATH "%u",
@@ -278,6 +215,10 @@ NULL
 };
 #endif
 
+/*
+ * Create the directory corresponding to dirname of `path' or fail.
+ */
+
 static void
 try_mkdir (const char *path)
 {
@@ -297,6 +238,14 @@ try_mkdir (const char *path)
     free (dir);
 }
 
+/*
+ * Allocate a display, returning the number of sockets in `number' and
+ * all the corresponding sockets in `sockets'.  If `tcp_socket' is
+ * true, also allcoaet a TCP socket.
+ *
+ * The return value is the display allocated or -1 if an error occurred.
+ */
+
 int
 get_xsockets (int *number, struct x_socket **sockets, int tcp_socket)
 {
@@ -381,7 +330,8 @@ get_xsockets (int *number, struct x_socket **sockets, int tcp_socket)
 }
 
 /*
- *
+ * Change owner on the `n' sockets in `sockets' to `uid', `gid'.
+ * Return 0 is succesful or -1 if an error occurred.
  */
 
 int
@@ -397,7 +347,8 @@ chown_xsockets (int n, struct x_socket *sockets, uid_t uid, gid_t gid)
 }
 
 /*
- *
+ * Connect to local display `dnr' with local transport.
+ * Return a file descriptor.
  */
 
 int
@@ -420,11 +371,18 @@ connect_local_xsocket (unsigned dnr)
      err (1, "connecting to local display %u", dnr);
 }
 
+/*
+ * Create a cookie file with a random cookie for the localhost.  The
+ * file name will be stored in `xauthfile' (but not larger than
+ * `xauthfile_size'), and the cookie returned in `cookie', `cookie_sz'.
+ * Return 0 if succesful, or errno.
+ */
+
 int
 create_and_write_cookie (char *xauthfile,
-			 size_t size,
+			 size_t xauthfile_size,
 			 u_char *cookie,
-			 size_t sz)
+			 size_t cookie_sz)
 {
      Xauth auth;
      char tmp[64];
@@ -432,7 +390,7 @@ create_and_write_cookie (char *xauthfile,
      FILE *f;
      char hostname[MaxHostNameLen];
      struct in_addr loopback;
-     struct hostent *h;
+     int saved_errno;
 
      gethostname (hostname, sizeof(hostname));
      loopback.s_addr = htonl(INADDR_LOOPBACK);
@@ -445,30 +403,27 @@ create_and_write_cookie (char *xauthfile,
      auth.number = tmp;
      auth.name = COOKIE_TYPE;
      auth.name_length = strlen(auth.name);
-     auth.data_length = sz;
+     auth.data_length = cookie_sz;
      auth.data = (char*)cookie;
-     des_rand_data (cookie, sz);
+     des_rand_data (cookie, cookie_sz);
 
-     strcpy_truncate(xauthfile, "/tmp/AXXXXXX", size);
+     strlcpy(xauthfile, "/tmp/AXXXXXX", xauthfile_size);
      fd = mkstemp(xauthfile);
      if(fd < 0) {
+	 saved_errno = errno;
 	 syslog(LOG_ERR, "create_and_write_cookie: mkstemp: %m");
-         return 1;
+         return saved_errno;
      }
      f = fdopen(fd, "r+");
      if(f == NULL){
+	 saved_errno = errno;
 	 close(fd);
-	 return 1;
+	 return errno;
      }
      if(XauWriteAuth(f, &auth) == 0) {
-	  fclose(f);
-	  return 1;
-     }
-
-     h = gethostbyname (hostname);
-     if (h == NULL) {
-	  fclose (f);
-	  return 1;
+	 saved_errno = errno;
+	 fclose(f);
+	 return saved_errno;
      }
 
      /*
@@ -486,12 +441,13 @@ create_and_write_cookie (char *xauthfile,
 #endif
 
      if (XauWriteAuth(f, &auth) == 0) {
-	  fclose (f);
-	  return 1;
+	 saved_errno = errno;
+	 fclose (f);
+	 return saved_errno;
      }
 
      if(fclose(f))
-	  return 1;
+	 return errno;
      return 0;
 }
 
@@ -542,9 +498,9 @@ verify_and_remove_cookies (int fd, int sock, int cookiesp)
 			   'b', 'a', 'd', ' ', 'c', 'o', 'o', 'k', 'i', 'e',
 			   0, 0};
 
-     if (krb_net_read (fd, beg, sizeof(beg)) != sizeof(beg))
+     if (net_read (fd, beg, sizeof(beg)) != sizeof(beg))
 	  return 1;
-     if (krb_net_write (sock, beg, 6) != 6)
+     if (net_write (sock, beg, 6) != 6)
 	  return 1;
      bigendianp = beg[0] == 'B';
      if (bigendianp) {
@@ -564,9 +520,9 @@ verify_and_remove_cookies (int fd, int sock, int cookiesp)
 	 free (protocol_name);
 	 return 1;
      }
-     if (krb_net_read (fd, protocol_name, n + npad) != n + npad)
+     if (net_read (fd, protocol_name, n + npad) != n + npad)
 	 goto fail;
-     if (krb_net_read (fd, protocol_data, d + dpad) != d + dpad)
+     if (net_read (fd, protocol_data, d + dpad) != d + dpad)
 	 goto fail;
      if (cookiesp) {
 	 if (strncmp (protocol_name, COOKIE_TYPE, strlen(COOKIE_TYPE)) != 0)
@@ -577,7 +533,7 @@ verify_and_remove_cookies (int fd, int sock, int cookiesp)
      }
      free (protocol_name);
      free (protocol_data);
-     if (krb_net_write (sock, zeros, 6) != 6)
+     if (net_write (sock, zeros, 6) != 6)
 	  return 1;
      return 0;
 refused:
@@ -590,13 +546,106 @@ refused:
      else
 	 refused[6] = 3;
 
-     krb_net_write (fd, refused, sizeof(refused));
+     net_write (fd, refused, sizeof(refused));
 fail:
      free (protocol_name);
      free (protocol_data);
      return 1;
 }
 
+/* 
+ * Return 0 iff `cookie' is compatible with the cookie for the
+ * localhost with name given in `ai' (or `hostname') and display
+ * number in `disp_nr'.
+ */
+
+static int
+match_local_auth (Xauth* auth,
+		  struct addrinfo *ai, const char *hostname, int disp_nr)
+{
+    int auth_disp;
+    char *tmp_disp;
+    struct addrinfo *a;
+    
+    tmp_disp = strndup (auth->number, auth->number_length);
+    if (tmp_disp == NULL)
+	return -1;
+    auth_disp = atoi(tmp_disp);
+    free (tmp_disp);
+    if (auth_disp != disp_nr)
+	return 1;
+    for (a = ai; a != NULL; a = a->ai_next) {
+	if ((auth->family == FamilyLocal
+	     || auth->family == FamilyWild)
+	    && a->ai_canonname != NULL
+	    && strncmp (auth->address,
+			a->ai_canonname,
+			auth->address_length) == 0)
+	    return 0;
+    }
+    if (hostname != NULL
+	&& (auth->family    == FamilyLocal
+	    || auth->family == FamilyWild)
+	&& strncmp (auth->address, hostname, auth->address_length) == 0)
+	return 0;
+    return 1;
+}
+
+/*
+ * Find `our' cookie from the cookie file `f' and return it or NULL.
+ */
+
+static Xauth*
+find_auth_cookie (FILE *f)
+{
+    Xauth *ret = NULL;
+    char local_hostname[MaxHostNameLen];
+    char *display = getenv("DISPLAY");
+    char d[MaxHostNameLen + 4];
+    char *colon;
+    struct addrinfo *ai;
+    struct addrinfo hints;
+    int disp;
+    int error;
+
+    if(display == NULL)
+	display = ":0";
+    strlcpy(d, display, sizeof(d));
+    display = d;
+    colon = strchr (display, ':');
+    if (colon == NULL)
+	disp = 0;
+    else {
+	*colon = '\0';
+	disp = atoi (colon + 1);
+    }
+    if (strcmp (display, "") == 0
+	|| strncmp (display, "unix", 4) == 0
+	|| strncmp (display, "localhost", 9) == 0) {
+	gethostname (local_hostname, sizeof(local_hostname));
+	display = local_hostname;
+    }
+    memset (&hints, 0, sizeof(hints));
+    hints.ai_flags    = AI_CANONNAME;
+    hints.ai_socktype = SOCK_STREAM;
+    hints.ai_protocol = IPPROTO_TCP;
+
+    error = getaddrinfo (display, NULL, &hints, &ai);
+    if (error)
+	ai = NULL;
+
+    for (; (ret = XauReadAuth (f)) != NULL; XauDisposeAuth(ret)) {
+	if (match_local_auth (ret, ai, display, disp) == 0) {
+	    if (ai != NULL)
+		freeaddrinfo (ai);
+	    return ret;
+	}
+    }
+    if (ai != NULL)
+	freeaddrinfo (ai);
+    return NULL;
+}
+
 /*
  * Get rid of the cookie that we were sent and get the correct one
  * from our own cookie file instead.
@@ -608,13 +657,12 @@ replace_cookie(int xserver, int fd, char *filename, int cookiesp) /* XXX */
      u_char beg[12];
      int bigendianp;
      unsigned n, d, npad, dpad;
-     Xauth *auth;
      FILE *f;
      u_char zeros[6] = {0, 0, 0, 0, 0, 0};
 
-     if (krb_net_read (fd, beg, sizeof(beg)) != sizeof(beg))
+     if (net_read (fd, beg, sizeof(beg)) != sizeof(beg))
 	  return 1;
-     if (krb_net_write (xserver, beg, 6) != 6)
+     if (net_write (xserver, beg, 6) != 6)
 	  return 1;
      bigendianp = beg[0] == 'B';
      if (bigendianp) {
@@ -627,52 +675,58 @@ replace_cookie(int xserver, int fd, char *filename, int cookiesp) /* XXX */
      if (n != 0 || d != 0)
 	  return 1;
      f = fopen(filename, "r");
-     if (f) {
-	  u_char len[6] = {0, 0, 0, 0, 0, 0};
+     if (f != NULL) {
+	 Xauth *auth = find_auth_cookie (f);
+	 u_char len[6] = {0, 0, 0, 0, 0, 0};
+	 
+	 fclose (f);
 
-	  auth = XauReadAuth(f);
-	  fclose(f);
-	  if (auth != NULL) {
-	      n = auth->name_length;
-	      d = auth->data_length;
-	  } else {
-	      n = 0;
-	      d = 0;
-	  }
-	  if (bigendianp) {
-	       len[0] = n >> 8;
-	       len[1] = n & 0xFF;
-	       len[2] = d >> 8;
-	       len[3] = d & 0xFF;
-	  } else {
-	       len[0] = n & 0xFF;
-	       len[1] = n >> 8;
-	       len[2] = d & 0xFF;
-	       len[3] = d >> 8;
-	  }
-	  if (krb_net_write (xserver, len, 6) != 6)
-	       return 1;
-	  if(n != 0 && krb_net_write (xserver, auth->name, n) != n)
-	       return 1;
-	  npad = (4 - (n % 4)) % 4;
-	  if (npad && krb_net_write (xserver, zeros, npad) != npad)
-	      return 1;
-	  if (d != 0 && krb_net_write (xserver, auth->data, d) != d)
-	       return 1;
-	  dpad = (4 - (d % 4)) % 4;
-	  if (dpad && krb_net_write (xserver, zeros, dpad) != dpad)
-	      return 1;
-	  XauDisposeAuth(auth);
+	 if (auth != NULL) {
+	     n = auth->name_length;
+	     d = auth->data_length;
+	 } else {
+	     n = 0;
+	     d = 0;
+	 }
+	 if (bigendianp) {
+	     len[0] = n >> 8;
+	     len[1] = n & 0xFF;
+	     len[2] = d >> 8;
+	     len[3] = d & 0xFF;
+	 } else {
+	     len[0] = n & 0xFF;
+	     len[1] = n >> 8;
+	     len[2] = d & 0xFF;
+	     len[3] = d >> 8;
+	 }
+	 if (net_write (xserver, len, 6) != 6) {
+	     XauDisposeAuth(auth);
+	     return 1;
+	 }
+	 if(n != 0 && net_write (xserver, auth->name, n) != n) {
+	     XauDisposeAuth(auth);
+	     return 1;
+	 }
+	 npad = (4 - (n % 4)) % 4;
+	 if (npad && net_write (xserver, zeros, npad) != npad) {
+	     XauDisposeAuth(auth);
+	     return 1;
+	 }
+	 if (d != 0 && net_write (xserver, auth->data, d) != d) {
+	     XauDisposeAuth(auth);
+	     return 1;
+	 }
+	 XauDisposeAuth(auth);
+	 dpad = (4 - (d % 4)) % 4;
+	 if (dpad && net_write (xserver, zeros, dpad) != dpad)
+	     return 1;
      } else {
-	  if(krb_net_write(xserver, zeros, 6) != 6)
-	       return 1;
+	 if(net_write(xserver, zeros, 6) != 6)
+	     return 1;
      }
      return 0;
 }
 
-
-
-
 /*
  * Some simple controls on the address and corresponding socket
  */
@@ -681,7 +735,7 @@ int
 suspicious_address (int sock, struct sockaddr_in addr)
 {
     char data[40];
-    int len = sizeof(data);
+    socklen_t len = sizeof(data);
 
     return addr.sin_addr.s_addr != htonl(INADDR_LOOPBACK)
 #if defined(IP_OPTIONS) && defined(HAVE_GETSOCKOPT)
@@ -690,3 +744,47 @@ suspicious_address (int sock, struct sockaddr_in addr)
 #endif
     ;
 }
+
+/*
+ * This really sucks, but these functions are used and if we're not
+ * linking against libkrb they don't exist.  Using the heimdal storage
+ * functions will not work either cause we do not always link with
+ * libkrb5 either.
+ */
+
+#ifndef KRB4
+
+int
+krb_get_int(void *f, u_int32_t *to, int size, int lsb)
+{
+    int i;
+    unsigned char *from = (unsigned char *)f;
+
+    *to = 0;
+    if(lsb){
+	for(i = size-1; i >= 0; i--)
+	    *to = (*to << 8) | from[i];
+    }else{
+	for(i = 0; i < size; i++)
+	    *to = (*to << 8) | from[i];
+    }
+    return size;
+}
+
+int
+krb_put_int(u_int32_t from, void *to, size_t rem, int size)
+{
+    int i;
+    unsigned char *p = (unsigned char *)to;
+
+    if (rem < size)
+	return -1;
+
+    for(i = size - 1; i >= 0; i--){
+	p[i] = from & 0xff;
+	from >>= 8;
+    }
+    return size;
+}
+
+#endif /* !KRB4 */
