@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997-1999 Kungliga Tekniska Hgskolan
+ * Copyright (c) 1997-2000 Kungliga Tekniska Hgskolan
  * (Royal Institute of Technology, Stockholm, Sweden). 
  * All rights reserved. 
  *
@@ -14,12 +14,7 @@
  *    notice, this list of conditions and the following disclaimer in the 
  *    documentation and/or other materials provided with the distribution. 
  *
- * 3. All advertising materials mentioning features or use of this software 
- *    must display the following acknowledgement: 
- *      This product includes software developed by Kungliga Tekniska 
- *      Hgskolan and its contributors. 
- *
- * 4. Neither the name of the Institute nor the names of its contributors 
+ * 3. Neither the name of the Institute nor the names of its contributors 
  *    may be used to endorse or promote products derived from this software 
  *    without specific prior written permission. 
  *
@@ -38,7 +33,13 @@
 
 #include "krb5_locl.h"
 
-RCSID("$Id: cache.c,v 1.38 1999/04/24 16:38:45 joda Exp $");
+RCSID("$Id: cache.c,v 1.45 2000/12/05 09:18:29 joda Exp $");
+
+/*
+ * Add a new ccache type with operations `ops', overwriting any
+ * existing one if `override'.
+ * Return an error code or 0.
+ */
 
 krb5_error_code
 krb5_cc_register(krb5_context context, 
@@ -46,28 +47,25 @@ krb5_cc_register(krb5_context context,
 		 krb5_boolean override)
 {
     int i;
-    if(context->cc_ops == NULL){
-	context->num_ops = 4;
-	context->cc_ops = calloc(context->num_ops, sizeof(*context->cc_ops));
-    }
-    for(i = 0; context->cc_ops[i].prefix && i < context->num_ops; i++){
-	if(strcmp(context->cc_ops[i].prefix, ops->prefix) == 0){
+
+    for(i = 0; i < context->num_cc_ops && context->cc_ops[i].prefix; i++) {
+	if(strcmp(context->cc_ops[i].prefix, ops->prefix) == 0) {
 	    if(override)
 		free(context->cc_ops[i].prefix);
 	    else
-		return KRB5_CC_TYPE_EXISTS; /* XXX */
+		return KRB5_CC_TYPE_EXISTS;
 	}
     }
-    if(i == context->num_ops){
-	krb5_cc_ops *o = realloc(context->cc_ops, 
-				 (context->num_ops + 4) * 
+    if(i == context->num_cc_ops) {
+	krb5_cc_ops *o = realloc(context->cc_ops,
+				 (context->num_cc_ops + 1) *
 				 sizeof(*context->cc_ops));
 	if(o == NULL)
 	    return KRB5_CC_NOMEM;
-	context->num_ops += 4;
+	context->num_cc_ops++;
 	context->cc_ops = o;
 	memset(context->cc_ops + i, 0, 
-	       (context->num_ops - i) * sizeof(*context->cc_ops));
+	       (context->num_cc_ops - i) * sizeof(*context->cc_ops));
     }
     memcpy(&context->cc_ops[i], ops, sizeof(context->cc_ops[i]));
     context->cc_ops[i].prefix = strdup(ops->prefix);
@@ -77,39 +75,66 @@ krb5_cc_register(krb5_context context,
     return 0;
 }
 
+/*
+ * Allocate memory for a new ccache in `id' with operations `ops'
+ * and name `residual'.
+ * Return 0 or an error code.
+ */
+
+static krb5_error_code
+allocate_ccache (krb5_context context,
+		 const krb5_cc_ops *ops,
+		 const char *residual,
+		 krb5_ccache *id)
+{
+    krb5_error_code ret;
+    krb5_ccache p;
+
+    p = malloc(sizeof(*p));
+    if(p == NULL)
+	return KRB5_CC_NOMEM;
+    p->ops = ops;
+    *id = p;
+    ret = p->ops->resolve(context, id, residual);
+    if(ret)
+	free(p);
+    return ret;
+}
+
+/*
+ * Find and allocate a ccache in `id' from the specification in `residual'.
+ * If the ccache name doesn't contain any colon, interpret it as a file name.
+ * Return 0 or an error code.
+ */
 
 krb5_error_code
 krb5_cc_resolve(krb5_context context,
-		const char *residual,
+		const char *name,
 		krb5_ccache *id)
 {
-    krb5_error_code ret;
     int i;
 
-    if(context->cc_ops == NULL){
-	ret = krb5_cc_register(context, &krb5_fcc_ops, 1);
-	if(ret) return ret;
-	ret = krb5_cc_register(context, &krb5_mcc_ops, 1);
-	if(ret) return ret;
-    }
+    for(i = 0; i < context->num_cc_ops && context->cc_ops[i].prefix; i++) {
+	size_t prefix_len = strlen(context->cc_ops[i].prefix);
 
-    for(i = 0; i < context->num_ops && context->cc_ops[i].prefix; i++)
-	if(strncmp(context->cc_ops[i].prefix, residual, 
-		   strlen(context->cc_ops[i].prefix)) == 0){
-	    krb5_ccache p;
-	    p = malloc(sizeof(*p));
-	    if(p == NULL)
-		return KRB5_CC_NOMEM;
-	    p->ops = &context->cc_ops[i];
-	    *id = p;
-	    ret =  p->ops->resolve(context, id, residual + 
-				   strlen(p->ops->prefix) + 1);
-	    if(ret) free(p);
-	    return ret;
+	if(strncmp(context->cc_ops[i].prefix, name, prefix_len) == 0
+	   && name[prefix_len] == ':') {
+	    return allocate_ccache (context, &context->cc_ops[i],
+				    name + prefix_len + 1,
+				    id);
 	}
-    return KRB5_CC_UNKNOWN_TYPE;
+    }
+    if (strchr (name, ':') == NULL)
+	return allocate_ccache (context, &krb5_fcc_ops, name, id);
+    else
+	return KRB5_CC_UNKNOWN_TYPE;
 }
 
+/*
+ * Generate a new ccache of type `ops' in `id'.
+ * Return 0 or an error code.
+ */
+
 krb5_error_code
 krb5_cc_gen_new(krb5_context context,
 		const krb5_cc_ops *ops,
@@ -120,11 +145,15 @@ krb5_cc_gen_new(krb5_context context,
     p = malloc (sizeof(*p));
     if (p == NULL)
 	return KRB5_CC_NOMEM;
-    p->ops = (krb5_cc_ops *)ops;
+    p->ops = ops;
     *id = p;
     return p->ops->gen_new(context, id);
 }
 
+/*
+ * Return the name of the ccache `id'
+ */
+
 const char*
 krb5_cc_get_name(krb5_context context,
 		 krb5_ccache id)
@@ -132,6 +161,10 @@ krb5_cc_get_name(krb5_context context,
     return id->ops->get_name(context, id);
 }
 
+/*
+ * Return the type of the ccache `id'.
+ */
+
 const char*
 krb5_cc_get_type(krb5_context context,
 		 krb5_ccache id)
@@ -139,16 +172,20 @@ krb5_cc_get_type(krb5_context context,
     return id->ops->prefix;
 }
 
+/*
+ * Return a pointer to a static string containing the default ccache name.
+ */
+
 const char*
 krb5_cc_default_name(krb5_context context)
 {
     static char name[1024];
     char *p;
+
     p = getenv("KRB5CCNAME");
-    if(p) {
-	strncpy (name, p, sizeof(name));
-	name[sizeof(name) - 1] = '\0';
-    } else
+    if(p)
+	strlcpy (name, p, sizeof(name));
+    else
 	snprintf(name,
 		 sizeof(name),
 		 "FILE:/tmp/krb5cc_%u",
@@ -156,8 +193,10 @@ krb5_cc_default_name(krb5_context context)
     return name;
 }
 
-
-
+/*
+ * Open the default ccache in `id'.
+ * Return 0 or an error code.
+ */
 
 krb5_error_code
 krb5_cc_default(krb5_context context,
@@ -168,6 +207,11 @@ krb5_cc_default(krb5_context context,
 			   id);
 }
 
+/*
+ * Create a new ccache in `id' for `primary_principal'.
+ * Return 0 or an error code.
+ */
+
 krb5_error_code
 krb5_cc_initialize(krb5_context context,
 		   krb5_ccache id,
@@ -177,6 +221,11 @@ krb5_cc_initialize(krb5_context context,
 }
 
 
+/*
+ * Remove the ccache `id'.
+ * Return 0 or an error code.
+ */
+
 krb5_error_code
 krb5_cc_destroy(krb5_context context,
 		krb5_ccache id)
@@ -188,6 +237,11 @@ krb5_cc_destroy(krb5_context context,
     return ret;
 }
 
+/*
+ * Stop using the ccache `id' and free the related resources.
+ * Return 0 or an error code.
+ */
+
 krb5_error_code
 krb5_cc_close(krb5_context context,
 	      krb5_ccache id)
@@ -198,6 +252,11 @@ krb5_cc_close(krb5_context context,
     return ret;
 }
 
+/*
+ * Store `creds' in the ccache `id'.
+ * Return 0 or an error code.
+ */
+
 krb5_error_code
 krb5_cc_store_cred(krb5_context context,
 		   krb5_ccache id,
@@ -206,11 +265,17 @@ krb5_cc_store_cred(krb5_context context,
     return id->ops->store(context, id, creds);
 }
 
+/*
+ * Retrieve the credential identified by `mcreds' (and `whichfields')
+ * from `id' in `creds'.
+ * Return 0 or an error code.
+ */
+
 krb5_error_code
 krb5_cc_retrieve_cred(krb5_context context,
 		      krb5_ccache id,
 		      krb5_flags whichfields,
-		      krb5_creds *mcreds,
+		      const krb5_creds *mcreds,
 		      krb5_creds *creds)
 {
     krb5_error_code ret;
@@ -227,6 +292,11 @@ krb5_cc_retrieve_cred(krb5_context context,
     return ret;
 }
 
+/*
+ * Return the principal of `id' in `principal'.
+ * Return 0 or an error code.
+ */
+
 krb5_error_code
 krb5_cc_get_principal(krb5_context context,
 		      krb5_ccache id,
@@ -235,40 +305,66 @@ krb5_cc_get_principal(krb5_context context,
     return id->ops->get_princ(context, id, principal);
 }
 
+/*
+ * Start iterating over `id', `cursor' is initialized to the
+ * beginning.
+ * Return 0 or an error code.
+ */
+
 krb5_error_code
 krb5_cc_start_seq_get (krb5_context context,
-		       krb5_ccache id,
+		       const krb5_ccache id,
 		       krb5_cc_cursor *cursor)
 {
     return id->ops->get_first(context, id, cursor);
 }
 
+/*
+ * Retrieve the next cred pointed to by (`id', `cursor') in `creds'
+ * and advance `cursor'.
+ * Return 0 or an error code.
+ */
+
 krb5_error_code
 krb5_cc_next_cred (krb5_context context,
-		   krb5_ccache id,
+		   const krb5_ccache id,
 		   krb5_creds *creds,
 		   krb5_cc_cursor *cursor)
 {
     return id->ops->get_next(context, id, cursor, creds);
 }
 
+/*
+ * Destroy the cursor `cursor'.
+ */
+
 krb5_error_code
 krb5_cc_end_seq_get (krb5_context context,
-		     krb5_ccache id,
+		     const krb5_ccache id,
 		     krb5_cc_cursor *cursor)
 {
     return id->ops->end_get(context, id, cursor);
 }
 
+/*
+ * Remove the credential identified by `cred', `which' from `id'.
+ */
+
 krb5_error_code
 krb5_cc_remove_cred(krb5_context context,
 		    krb5_ccache id,
 		    krb5_flags which,
 		    krb5_creds *cred)
 {
-    return id->ops->remove_cred(context, id, which, cred);
+    if(id->ops->remove_cred == NULL)
+	return EACCES; /* XXX */
+    return (*id->ops->remove_cred)(context, id, which, cred);
 }
 
+/*
+ * Set the flags of `id' to `flags'.
+ */
+
 krb5_error_code
 krb5_cc_set_flags(krb5_context context,
 		  krb5_ccache id,
@@ -277,15 +373,20 @@ krb5_cc_set_flags(krb5_context context,
     return id->ops->set_flags(context, id, flags);
 }
 		    
+/*
+ * Copy the contents of `from' to `to'.
+ */
+
 krb5_error_code
 krb5_cc_copy_cache(krb5_context context,
-		   krb5_ccache from,
+		   const krb5_ccache from,
 		   krb5_ccache to)
 {
     krb5_error_code ret;
     krb5_cc_cursor cursor;
     krb5_creds cred;
     krb5_principal princ;
+
     ret = krb5_cc_get_principal(context, from, &princ);
     if(ret)
 	return ret;
@@ -308,9 +409,13 @@ krb5_cc_copy_cache(krb5_context context,
     return ret;
 }
 
+/*
+ * Return the version of `id'.
+ */
+
 krb5_error_code
 krb5_cc_get_version(krb5_context context,
-		    krb5_ccache id)
+		    const krb5_ccache id)
 {
     if(id->ops->get_version)
 	return id->ops->get_version(context, id);
