@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997, 1998, 1999 Kungliga Tekniska Hgskolan
+ * Copyright (c) 1997 - 2000 Kungliga Tekniska Hgskolan
  * (Royal Institute of Technology, Stockholm, Sweden). 
  * All rights reserved. 
  *
@@ -14,12 +14,7 @@
  *    notice, this list of conditions and the following disclaimer in the 
  *    documentation and/or other materials provided with the distribution. 
  *
- * 3. All advertising materials mentioning features or use of this software 
- *    must display the following acknowledgement: 
- *      This product includes software developed by Kungliga Tekniska 
- *      Hgskolan and its contributors. 
- *
- * 4. Neither the name of the Institute nor the names of its contributors 
+ * 3. Neither the name of the Institute nor the names of its contributors 
  *    may be used to endorse or promote products derived from this software 
  *    without specific prior written permission. 
  *
@@ -38,7 +33,7 @@
 
 #include "krb5_locl.h"
 
-RCSID("$Id: init_creds_pw.c,v 1.31 1999/04/20 22:11:54 assar Exp $");
+RCSID("$Id: init_creds_pw.c,v 1.44 2000/07/24 03:46:40 assar Exp $");
 
 static int
 get_config_time (krb5_context context,
@@ -93,7 +88,7 @@ init_cred (krb5_context context,
     krb5_error_code ret;
     krb5_realm *client_realm;
     int tmp;
-    int32_t now;
+    krb5_timestamp now;
 
     krb5_timeofday (context, &now);
 
@@ -169,7 +164,7 @@ print_expire (krb5_context context,
 {
     int i;
     LastReq *lr = &rep->enc_part.last_req;
-    int32_t sec;
+    krb5_timestamp sec;
     time_t t;
 
     krb5_timeofday (context, &sec);
@@ -183,9 +178,9 @@ print_expire (krb5_context context,
 	if (lr->val[i].lr_type == 6
 	    && lr->val[i].lr_value <= t) {
 	    char *p;
+	    time_t tmp = lr->val[i].lr_value;
 	    
-	    asprintf (&p, "Your password will expire at %s",
-		      ctime(&lr->val[i].lr_value));
+	    asprintf (&p, "Your password will expire at %s", ctime(&tmp));
 	    (*prompter) (context, data, p, 0, NULL);
 	    free (p);
 	    return;
@@ -195,9 +190,9 @@ print_expire (krb5_context context,
     if (rep->enc_part.key_expiration
 	&& *rep->enc_part.key_expiration <= t) {
 	char *p;
+	time_t t = *rep->enc_part.key_expiration;
 
-	asprintf (&p, "Your password/account will expire at %s",
-		  ctime(rep->enc_part.key_expiration));
+	asprintf (&p, "Your password/account will expire at %s", ctime(&t));
 	(*prompter) (context, data, p, 0, NULL);
 	free (p);
     }
@@ -268,6 +263,8 @@ get_init_creds_common(krb5_context context,
     }
     if (options->flags & KRB5_GET_INIT_CREDS_OPT_SALT)
 	;			/* XXX */
+    if (options->flags & KRB5_GET_INIT_CREDS_OPT_ANONYMOUS)
+	flags->b.request_anonymous = options->anonymous;
     return 0;
 }
 
@@ -296,9 +293,12 @@ change_password (krb5_context context,
 
     krb5_get_init_creds_opt_init (&options);
     krb5_get_init_creds_opt_set_tkt_life (&options, 60);
-    krb5_get_init_creds_opt_set_preauth_list (&options,
-					      old_options->preauth_list,
-					      old_options->preauth_list_length);					      
+    krb5_get_init_creds_opt_set_forwardable (&options, FALSE);
+    krb5_get_init_creds_opt_set_proxiable (&options, FALSE);
+    if (old_options->flags & KRB5_GET_INIT_CREDS_OPT_PREAUTH_LIST)
+	krb5_get_init_creds_opt_set_preauth_list (&options,
+						  old_options->preauth_list,
+						  old_options->preauth_list_length);					      
 
     krb5_data_zero (&result_code_string);
     krb5_data_zero (&result_string);
@@ -320,7 +320,7 @@ change_password (krb5_context context,
 	password_data.length = sizeof(buf1);
 
 	prompt.hidden = 1;
-	prompt.prompt = "New password";
+	prompt.prompt = "New password: ";
 	prompt.reply  = &password_data;
 
 	ret = (*prompter) (context, data, "Changing password", 1, &prompt);
@@ -331,7 +331,7 @@ change_password (krb5_context context,
 	password_data.length = sizeof(buf2);
 
 	prompt.hidden = 1;
-	prompt.prompt = "Repeat new password";
+	prompt.prompt = "Repeat new password: ";
 	prompt.reply  = &password_data;
 
 	ret = (*prompter) (context, data, "Changing password", 1, &prompt);
@@ -358,7 +358,7 @@ change_password (krb5_context context,
     ret = (*prompter) (context, data, p, 0, NULL);
     free (p);
     if (result_code == 0) {
-	strncpy (newpw, buf1, newpw_sz);
+	strlcpy (newpw, buf1, newpw_sz);
 	ret = 0;
     } else
 	ret = ENOTTY;
@@ -388,7 +388,6 @@ krb5_get_init_creds_password(krb5_context context,
     krb5_addresses *addrs = NULL;
     krb5_enctype *etypes = NULL;
     krb5_preauthtype *pre_auth_types = NULL;
-    krb5_preauthdata *preauth = NULL, preauth2;
     krb5_creds this_cred;
     krb5_kdc_rep kdc_reply;
     char buf[BUFSIZ];
@@ -418,7 +417,7 @@ krb5_get_init_creds_password(krb5_context context,
 	free (prompt.prompt);
 	if (ret) {
 	    memset (buf, 0, sizeof(buf));
-	    ret = KRB5KDC_ERR_NONE;
+	    ret = KRB5_LIBOS_PWDINTR;
 	    goto out;
 	}
 	password = password_data.data;
@@ -432,7 +431,7 @@ krb5_get_init_creds_password(krb5_context context,
 				addrs,
 				etypes,
 				pre_auth_types,
-				preauth,
+				NULL,
 				krb5_password_key_proc,
 				password,
 				NULL,
@@ -444,6 +443,12 @@ krb5_get_init_creds_password(krb5_context context,
 	    done = 1;
 	    break;
 	case KRB5KDC_ERR_KEY_EXPIRED :
+	    /* try to avoid recursion */
+
+	    if (in_tkt_service != NULL
+		&& strcmp (in_tkt_service, "kadmin/changepw") == 0)
+		goto out;
+
 	    ret = change_password (context,
 				   client,
 				   password,
@@ -456,52 +461,6 @@ krb5_get_init_creds_password(krb5_context context,
 		goto out;
 	    password = buf;
 	    break;
-	case KRB5KDC_ERR_PREAUTH_REQUIRED : {
-	    if(kdc_reply.error.e_data) {
-		METHOD_DATA md;
-		int i;
-		decode_METHOD_DATA(kdc_reply.error.e_data->data, 
-				   kdc_reply.error.e_data->length, 
-				   &md, 
-				   NULL);
-		for(i = 0; i < md.len; i++){
-		    switch(md.val[i].padata_type){
-		    case pa_enc_timestamp:
-			if (pre_auth_types)
-			    free (pre_auth_types);
-			ALLOC(pre_auth_types, 2);
-			if (pre_auth_types == NULL) {
-			    free_METHOD_DATA(&md);
-			    goto out;
-			}
-			pre_auth_types[0] = KRB5_PADATA_ENC_TIMESTAMP;
-			pre_auth_types[1] = KRB5_PADATA_NONE;
-			break;
-		    case pa_etype_info:
-			preauth = &preauth2;
-			ALLOC_SEQ(preauth, 1);
-			preauth->val[0].type = KRB5_PADATA_ENC_TIMESTAMP;
-			krb5_decode_ETYPE_INFO(context,
-					       md.val[i].padata_value.data, 
-					       md.val[i].padata_value.length,
-					       &preauth->val[0].info,
-					       NULL);
-			break;
-		    }
-		}
-		free_METHOD_DATA(&md);
-	    } else {
-		if (pre_auth_types)
-		    free (pre_auth_types);
-		pre_auth_types = malloc(2 * sizeof(*pre_auth_types));
-		if (pre_auth_types == NULL)
-		    goto out;
-		pre_auth_types[0] = KRB5_PADATA_ENC_TIMESTAMP;
-		pre_auth_types[1] = KRB5_PADATA_NONE;
-	    }
-	    krb5_free_kdc_rep (context, &kdc_reply);
-	    break;
-	}
 	default:
 	    goto out;
 	}
@@ -519,10 +478,6 @@ out:
 	krb5_free_kdc_rep (context, &kdc_reply);
 
     free (pre_auth_types);
-    if(preauth) {
-	free_ETYPE_INFO(&preauth->val[0].info);
-	free(preauth->val);
-    }
     free (etypes);
     if (ret == 0 && creds)
 	*creds = this_cred;
