@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997 Kungliga Tekniska Hgskolan
+ * Copyright (c) 1997-2000 Kungliga Tekniska Hgskolan
  * (Royal Institute of Technology, Stockholm, Sweden). 
  * All rights reserved. 
  *
@@ -14,12 +14,7 @@
  *    notice, this list of conditions and the following disclaimer in the 
  *    documentation and/or other materials provided with the distribution. 
  *
- * 3. All advertising materials mentioning features or use of this software 
- *    must display the following acknowledgement: 
- *      This product includes software developed by Kungliga Tekniska 
- *      Hgskolan and its contributors. 
- *
- * 4. Neither the name of the Institute nor the names of its contributors 
+ * 3. Neither the name of the Institute nor the names of its contributors 
  *    may be used to endorse or promote products derived from this software 
  *    without specific prior written permission. 
  *
@@ -37,10 +32,16 @@
  */
 
 #include "krb5_locl.h"
+#ifdef HAVE_RES_SEARCH
+#define USE_RESOLVER
+#endif
+#ifdef HAVE_ARPA_NAMESER_H
+#include <arpa/nameser.h>
+#endif
+#include <fnmatch.h>
+#include "resolve.h"
 
-RCSID("$Id: principal.c,v 1.28 1997/08/13 22:20:57 joda Exp $");
-
-/* Public principal handling functions */
+RCSID("$Id: principal.c,v 1.73 2000/10/16 03:42:14 assar Exp $");
 
 #define princ_num_comp(P) ((P)->name.name_string.len)
 #define princ_type(P) ((P)->name.name_type)
@@ -63,7 +64,7 @@ krb5_parse_name(krb5_context context,
 		const char *name,
 		krb5_principal *principal)
 {
-
+    krb5_error_code ret;
     general_string *comp;
     general_string realm;
     int ncomp;
@@ -88,9 +89,15 @@ krb5_parse_name(krb5_context context,
 	    ncomp++;
     }
     comp = calloc(ncomp, sizeof(*comp));
+    if (comp == NULL)
+	return ENOMEM;
   
     n = 0;
     start = q = p = s = strdup(name);
+    if (start == NULL) {
+	free (comp);
+	return ENOMEM;
+    }
     while(*p){
 	c = *p++;
 	if(c == '\\'){
@@ -105,16 +112,15 @@ krb5_parse_name(krb5_context context,
 		c = '\0';
 	}else if(c == '/' || c == '@'){
 	    if(got_realm){
-	    exit:
-		while(n>0){
-		    free(comp[--n]);
-		}
-		free(comp);
-		free(s);
-		return KRB5_PARSE_MALFORMED;
+		ret = KRB5_PARSE_MALFORMED;
+		goto exit;
 	    }else{
 		comp[n] = malloc(q - start + 1);
-		strncpy(comp[n], start, q - start);
+		if (comp[n] == NULL) {
+		    ret = ENOMEM;
+		    goto exit;
+		}
+		memcpy(comp[n], start, q - start);
 		comp[n][q - start] = 0;
 		n++;
 	    }
@@ -123,101 +129,180 @@ krb5_parse_name(krb5_context context,
 	    start = q;
 	    continue;
 	}
-	if(got_realm && (c == ':' || c == '/' || c == '\0'))
+	if(got_realm && (c == ':' || c == '/' || c == '\0')) {
+	    ret = KRB5_PARSE_MALFORMED;
 	    goto exit;
+	}
 	*q++ = c;
     }
     if(got_realm){
 	realm = malloc(q - start + 1);
-	strncpy(realm, start, q - start);
+	if (realm == NULL) {
+	    ret = ENOMEM;
+	    goto exit;
+	}
+	memcpy(realm, start, q - start);
 	realm[q - start] = 0;
     }else{
-	krb5_get_default_realm (context, &realm);
+	ret = krb5_get_default_realm (context, &realm);
+	if (ret)
+	    goto exit;
 
 	comp[n] = malloc(q - start + 1);
-	strncpy(comp[n], start, q - start);
+	if (comp[n] == NULL) {
+	    ret = ENOMEM;
+	    goto exit;
+	}
+	memcpy(comp[n], start, q - start);
 	comp[n][q - start] = 0;
 	n++;
     }
     *principal = malloc(sizeof(**principal));
+    if (*principal == NULL) {
+	ret = ENOMEM;
+	goto exit;
+    }
     (*principal)->name.name_type = KRB5_NT_PRINCIPAL;
     (*principal)->name.name_string.val = comp;
     princ_num_comp(*principal) = n;
     (*principal)->realm = realm;
     free(s);
     return 0;
+exit:
+    while(n>0){
+	free(comp[--n]);
+    }
+    free(comp);
+    free(s);
+    return ret;
 }
 
+static const char quotable_chars[] = " \n\t\b\\/@";
+static const char replace_chars[] = " ntb\\/@";
+
+#define add_char(BASE, INDEX, LEN, C) do { if((INDEX) < (LEN)) (BASE)[(INDEX)++] = (C); }while(0);
+
 static size_t
-quote_string(char *s, char **out)
+quote_string(const char *s, char *out, size_t index, size_t len)
 {
-    size_t len;
-    char *p;
-    char *tmp;
-    char c = 0;
-    tmp = *out;
-    if(tmp)
-	len = strlen(tmp);
-    else 
-	len = 0;
-    for(p = s; *p; p++){
-	if(*p == '\n')
-	    c = 'n';
-	else if(*p == '\t')
-	    c = 't';
-	else if(*p == '\b')
-	    c = 'b';
-	else if(*p == '\0')
-	    c = '0';
-	else if(*p == '/')
-	    c='/';      
-	else if(*p == '@')
-	    c = '@';
-	if(c){
-	    tmp = realloc(tmp, len + 2);
-	    tmp[len] = '\\';
-	    tmp[len + 1] = c;
-	    len += 2;
-	    c = 0;
-	}else{
-	    tmp = realloc(tmp, len + 1);
-	    tmp[len] = *p;
-	    len++;
-	}
+    const char *p, *q;
+    for(p = s; *p && index < len; p++){
+	if((q = strchr(quotable_chars, *p))){
+	    add_char(out, index, len, '\\');
+	    add_char(out, index, len, replace_chars[q - quotable_chars]);
+	}else
+	    add_char(out, index, len, *p);
     }
-    tmp = realloc(tmp, len + 1);
-    tmp[len] = 0;
-    *out = tmp;
-    return len;
+    if(index < len)
+	out[index] = '\0';
+    return index;
 }
 
 
-krb5_error_code
-krb5_unparse_name(krb5_context context,
-		  krb5_principal principal,
-		  char **name)
+static krb5_error_code
+unparse_name_fixed(krb5_context context,
+		   krb5_const_principal principal,
+		   char *name,
+		   size_t len,
+		   krb5_boolean short_form)
 {
-    size_t len;
-    char *s;
+    size_t index = 0;
     int i;
-    int ncomp = princ_num_comp(principal);
-    s = NULL;
-    for (i = 0; i < ncomp; i++){
-	if(i){
-	    s = realloc(s, len + 2);
-	    s[len] = '/';
-	    s[len + 1] = 0;
-	}
-	len = quote_string(princ_ncomp(principal, i), &s);
+    for(i = 0; i < princ_num_comp(principal); i++){
+	if(i)
+	    add_char(name, index, len, '/');
+	index = quote_string(princ_ncomp(principal, i), name, index, len);
+	if(index == len)
+	    return ERANGE;
+    } 
+    /* add realm if different from default realm */
+    if(short_form) {
+	krb5_realm r;
+	krb5_error_code ret;
+	ret = krb5_get_default_realm(context, &r);
+	if(ret)
+	    return ret;
+	if(strcmp(princ_realm(principal), r) != 0)
+	    short_form = 0;
+	free(r);
+    }
+    if(!short_form) {
+	add_char(name, index, len, '@');
+	index = quote_string(princ_realm(principal), name, index, len);
+	if(index == len)
+	    return ERANGE;
     }
-    s = realloc(s, len + 2);
-    s[len] = '@';
-    s[len + 1] = 0;
-    len = quote_string(princ_realm(principal), &s);
-    *name = s;
     return 0;
 }
 
+krb5_error_code
+krb5_unparse_name_fixed(krb5_context context,
+			krb5_const_principal principal,
+			char *name,
+			size_t len)
+{
+    return unparse_name_fixed(context, principal, name, len, FALSE);
+}
+
+krb5_error_code
+krb5_unparse_name_fixed_short(krb5_context context,
+			      krb5_const_principal principal,
+			      char *name,
+			      size_t len)
+{
+    return unparse_name_fixed(context, principal, name, len, TRUE);
+}
+
+static krb5_error_code
+unparse_name(krb5_context context,
+	     krb5_const_principal principal,
+	     char **name,
+	     krb5_boolean short_flag)
+{
+    size_t len = 0, plen;
+    int i;
+    krb5_error_code ret;
+    /* count length */
+    plen = strlen(princ_realm(principal));
+    if(strcspn(princ_realm(principal), quotable_chars) == plen)
+	len += plen;
+    else
+	len += 2*plen;
+    len++;
+    for(i = 0; i < princ_num_comp(principal); i++){
+	plen = strlen(princ_ncomp(principal, i));
+	if(strcspn(princ_ncomp(principal, i), quotable_chars) == plen)
+	    len += plen;
+	else
+	    len += 2*plen;
+	len++;
+    }
+    *name = malloc(len);
+    if(len != 0 && *name == NULL)
+	return ENOMEM;
+    ret = unparse_name_fixed(context, principal, *name, len, short_flag);
+    if(ret)
+	free(*name);
+    return ret;
+}
+
+krb5_error_code
+krb5_unparse_name(krb5_context context,
+		  krb5_const_principal principal,
+		  char **name)
+{
+    return unparse_name(context, principal, name, FALSE);
+}
+
+krb5_error_code
+krb5_unparse_name_short(krb5_context context,
+			krb5_const_principal principal,
+			char **name)
+{
+    return unparse_name(context, principal, name, TRUE);
+}
+
+#if 0 /* not implemented */
 
 krb5_error_code
 krb5_unparse_name_ext(krb5_context context,
@@ -225,10 +310,10 @@ krb5_unparse_name_ext(krb5_context context,
 		      char **name,
 		      size_t *size)
 {
-    fprintf(stderr, "krb5_unparse_name_ext: not implemented\n");
-    abort();
+    krb5_abortx(context, "unimplemented krb5_unparse_name_ext called");
 }
 
+#endif
 
 krb5_realm*
 krb5_princ_realm(krb5_context context,
@@ -264,15 +349,21 @@ krb5_build_principal(krb5_context context,
 
 static krb5_error_code
 append_component(krb5_context context, krb5_principal p, 
-		 general_string comp)
+		 const char *comp,
+		 size_t comp_len)
 {
     general_string *tmp;
     size_t len = princ_num_comp(p);
+
     tmp = realloc(princ_comp(p), (len + 1) * sizeof(*tmp));
     if(tmp == NULL)
 	return ENOMEM;
     princ_comp(p) = tmp;
-    princ_ncomp(p, len) = strdup(comp);
+    princ_ncomp(p, len) = malloc(comp_len + 1);
+    if (princ_ncomp(p, len) == NULL)
+	return ENOMEM;
+    memcpy (princ_ncomp(p, len), comp, comp_len);
+    princ_ncomp(p, len)[comp_len] = '\0';
     princ_num_comp(p)++;
     return 0;
 }
@@ -281,13 +372,13 @@ static void
 va_ext_princ(krb5_context context, krb5_principal p, va_list ap)
 {
     while(1){
-	char *s;
+	const char *s;
 	int len;
 	len = va_arg(ap, int);
 	if(len == 0)
 	    break;
-	s = va_arg(ap, char*);
-	append_component(context, p, s);
+	s = va_arg(ap, const char*);
+	append_component(context, p, s, len);
     }
 }
 
@@ -295,11 +386,11 @@ static void
 va_princ(krb5_context context, krb5_principal p, va_list ap)
 {
     while(1){
-	char *s;
-	s = va_arg(ap, char*);
+	const char *s;
+	s = va_arg(ap, const char*);
 	if(s == NULL)
 	    break;
-	append_component(context, p, s);
+	append_component(context, p, s, strlen(s));
     }
 }
 
@@ -313,7 +404,6 @@ build_principal(krb5_context context,
 		va_list ap)
 {
     krb5_principal p;
-    int n;
   
     p = calloc(1, sizeof(*p));
     if (p == NULL)
@@ -340,7 +430,7 @@ krb5_make_principal(krb5_context context,
     krb5_error_code ret;
     krb5_realm r = NULL;
     va_list ap;
-    if(realm == NULL){
+    if(realm == NULL) {
 	ret = krb5_get_default_realm(context, &r);
 	if(ret)
 	    return ret;
@@ -399,11 +489,15 @@ krb5_copy_principal(krb5_context context,
     krb5_principal p = malloc(sizeof(*p));
     if (p == NULL)
 	return ENOMEM;
-    copy_Principal(inprinc, p);
+    if(copy_Principal(inprinc, p))
+	return ENOMEM;
     *outprinc = p;
     return 0;
 }
 
+/*
+ * return TRUE iff princ1 == princ2 (without considering the realm)
+ */
 
 krb5_boolean
 krb5_principal_compare_any_realm(krb5_context context,
@@ -420,6 +514,10 @@ krb5_principal_compare_any_realm(krb5_context context,
     return TRUE;
 }
 
+/*
+ * return TRUE iff princ1 == princ2
+ */
+
 krb5_boolean
 krb5_principal_compare(krb5_context context,
 		       krb5_const_principal princ1,
@@ -430,6 +528,9 @@ krb5_principal_compare(krb5_context context,
     return krb5_principal_compare_any_realm(context, princ1, princ2);
 }
 
+/*
+ * return TRUE iff realm(princ1) == realm(princ2)
+ */
 
 krb5_boolean
 krb5_realm_compare(krb5_context context,
@@ -439,16 +540,103 @@ krb5_realm_compare(krb5_context context,
     return strcmp(princ_realm(princ1), princ_realm(princ2)) == 0;
 }
 
+/*
+ * return TRUE iff princ matches pattern
+ */
+
+krb5_boolean
+krb5_principal_match(krb5_context context,
+		     krb5_const_principal princ,
+		     krb5_const_principal pattern)
+{
+    int i;
+    if(princ_num_comp(princ) != princ_num_comp(pattern))
+	return FALSE;
+    if(fnmatch(princ_realm(pattern), princ_realm(princ), 0) != 0)
+	return FALSE;
+    for(i = 0; i < princ_num_comp(princ); i++){
+	if(fnmatch(princ_ncomp(pattern, i), princ_ncomp(princ, i), 0) != 0)
+	    return FALSE;
+    }
+    return TRUE;
+}
+
+
+struct v4_name_convert {
+    const char *from;
+    const char *to; 
+} default_v4_name_convert[] = {
+    { "ftp",	"ftp" },
+    { "hprop",	"hprop" },
+    { "pop",	"pop" },
+    { "imap",	"imap" },
+    { "rcmd",	"host" },
+    { NULL, NULL }
+};
+
+/*
+ * return the converted instance name of `name' in `realm'.
+ * look in the configuration file and then in the default set above.
+ * return NULL if no conversion is appropriate.
+ */
+
+static const char*
+get_name_conversion(krb5_context context, const char *realm, const char *name)
+{
+    struct v4_name_convert *q;
+    const char *p;
+
+    p = krb5_config_get_string(context, NULL, "realms", realm,
+			       "v4_name_convert", "host", name, NULL);
+    if(p == NULL)
+	p = krb5_config_get_string(context, NULL, "libdefaults", 
+				   "v4_name_convert", "host", name, NULL);
+    if(p)
+	return p;
+
+    /* XXX should be possible to override default list */
+    p = krb5_config_get_string(context, NULL,
+			       "realms",
+			       realm,
+			       "v4_name_convert",
+			       "plain",
+			       name,
+			       NULL);
+    if(p)
+	return NULL;
+    p = krb5_config_get_string(context, NULL,
+			       "libdefaults",
+			       "v4_name_convert",
+			       "plain",
+			       name,
+			       NULL);
+    if(p)
+	return NULL;
+    for(q = default_v4_name_convert; q->from; q++)
+	if(strcmp(q->from, name) == 0)
+	    return q->to;
+    return NULL;
+}
+
+/*
+ * convert the v4 principal `name.instance@realm' to a v5 principal in `princ'.
+ * if `resolve', use DNS.
+ * if `func', use that function for validating the conversion
+ */
+
 krb5_error_code
-krb5_425_conv_principal(krb5_context context,
-			const char *name,
-			const char *instance,
-			const char *realm,
-			krb5_principal *princ)
+krb5_425_conv_principal_ext(krb5_context context,
+			    const char *name,
+			    const char *instance,
+			    const char *realm,
+			    krb5_boolean (*func)(krb5_context, krb5_principal),
+			    krb5_boolean resolve,
+			    krb5_principal *princ)
 {
     const char *p;
     krb5_error_code ret;
-    char host[128];
+    krb5_principal pr;
+    char host[MAXHOSTNAMELEN];
 
     /* do the following: if the name is found in the
        `v4_name_convert:host' part, is is assumed to be a `host' type
@@ -464,41 +652,100 @@ krb5_425_conv_principal(krb5_context context,
 	instance = NULL;
 	goto no_host;
     }
-    p = krb5_config_get_string(context->cf, "realms", realm,
-			       "v4_name_convert", "host", name, NULL);
+    p = get_name_conversion(context, realm, name);
     if(p == NULL)
-	p = krb5_config_get_string(context->cf, "libdefaults", 
-				   "v4_name_convert", "host", name, NULL);
+	goto no_host;
+    name = p;
+    p = krb5_config_get_string(context, NULL, "realms", realm, 
+			       "v4_instance_convert", instance, NULL);
     if(p){
-	name = p;
-	p = krb5_config_get_string(context->cf, "realms", realm, 
-				   "v4_instance_convert", instance, NULL);
-	if(p){
-	    instance = p;
-	    goto done;
+	instance = p;
+	ret = krb5_make_principal(context, &pr, realm, name, instance, NULL);
+	if(func == NULL || (*func)(context, pr)){
+	    *princ = pr;
+	    return 0;
 	}
-	if(krb5_config_get_bool(context->cf, "libdefaults", 
-				"v4_instance_resolve", NULL)){
-	    struct hostent *hp = gethostbyname(instance);
-	    if(hp){
-		instance = hp->h_name;
-		goto done;
+	krb5_free_principal(context, pr);
+	*princ = NULL;
+	return HEIM_ERR_V4_PRINC_NO_CONV;
+    }
+    if(resolve){
+	const char *inst = NULL;
+#ifdef USE_RESOLVER
+	struct dns_reply *r;
+	r = dns_lookup(instance, "a");
+	if(r && r->head && r->head->type == T_A)
+	    inst = r->head->domain;
+#else
+	struct hostent *hp = roken_gethostbyname(instance);
+	if(hp)
+	    inst = hp->h_name;
+#endif
+	if(inst) {
+	    char *low_inst = strdup(inst);
+
+	    if (low_inst == NULL) {
+#ifdef USE_RESOLVER
+		dns_free_data(r);
+#endif
+		return ENOMEM;
+	    }
+	    ret = krb5_make_principal(context, &pr, realm, name, low_inst,
+				      NULL);
+	    free (low_inst);
+	    if(ret == 0) {
+		if(func == NULL || (*func)(context, pr)){
+		    *princ = pr;
+#ifdef USE_RESOLVER
+		    dns_free_data(r);
+#endif
+		    return 0;
+		}
+		krb5_free_principal(context, pr);
 	    }
 	}
-	p = krb5_config_get_string(context->cf, "realms", realm, 
-				   "default_domain", NULL);
-	if(p == NULL){
-	    /* should this be an error or should it silently
-	       succeed? */
-	    return HEIM_ERR_V4_PRINC_NO_CONV;
-	}
-	
-	snprintf(host, sizeof(host), "%s.%s", instance, p);
-	instance = host;
-	goto done;
+#ifdef USE_RESOLVER
+	if(r) 
+	    dns_free_data(r);
+#endif
     }
+    {
+	char **domains, **d;
+	domains = krb5_config_get_strings(context, NULL, "realms", realm,
+					  "v4_domains", NULL);
+	for(d = domains; d && *d; d++){
+	    snprintf(host, sizeof(host), "%s.%s", instance, *d);
+	    ret = krb5_make_principal(context, &pr, realm, name, host, NULL);
+	    if(func == NULL || (*func)(context, pr)){
+		*princ = pr;
+		krb5_config_free_strings(domains);
+		return 0;
+	    }
+	    krb5_free_principal(context, pr);
+	}
+	krb5_config_free_strings(domains);
+    }
+    
+    
+    p = krb5_config_get_string(context, NULL, "realms", realm, 
+			       "default_domain", NULL);
+    if(p == NULL){
+	/* this should be an error, just faking a name is not good */
+	return HEIM_ERR_V4_PRINC_NO_CONV;
+    }
+	
+    if (*p == '.')
+	++p;
+    snprintf(host, sizeof(host), "%s.%s", instance, p);
+    ret = krb5_make_principal(context, &pr, realm, name, host, NULL);
+    if(func == NULL || (*func)(context, pr)){
+	*princ = pr;
+	return 0;
+    }
+    krb5_free_principal(context, pr);
+    return HEIM_ERR_V4_PRINC_NO_CONV;
 no_host:
-    p = krb5_config_get_string(context->cf,
+    p = krb5_config_get_string(context, NULL,
 			       "realms",
 			       realm,
 			       "v4_name_convert",
@@ -506,7 +753,7 @@ no_host:
 			       name,
 			       NULL);
     if(p == NULL)
-	p = krb5_config_get_string(context->cf,
+	p = krb5_config_get_string(context, NULL,
 				   "libdefaults",
 				   "v4_name_convert",
 				   "plain",
@@ -515,38 +762,108 @@ no_host:
     if(p)
 	name = p;
     
-done:		
-    return krb5_make_principal(context, princ, realm, name, instance, NULL);
+    ret = krb5_make_principal(context, &pr, realm, name, instance, NULL);
+    if(func == NULL || (*func)(context, pr)){
+	*princ = pr;
+	return 0;
+    }
+    krb5_free_principal(context, pr);
+    return HEIM_ERR_V4_PRINC_NO_CONV;
 }
 
-static char*
+krb5_error_code
+krb5_425_conv_principal(krb5_context context,
+			const char *name,
+			const char *instance,
+			const char *realm,
+			krb5_principal *princ)
+{
+    krb5_boolean resolve = krb5_config_get_bool(context,
+						NULL,
+						"libdefaults", 
+						"v4_instance_resolve", 
+						NULL);
+
+    return krb5_425_conv_principal_ext(context, name, instance, realm, 
+				       NULL, resolve, princ);
+}
+
+
+static int
+check_list(const krb5_config_binding *l, const char *name, const char **out)
+{
+    while(l){
+	if (l->type != krb5_config_string)
+	    continue;
+	if(strcmp(name, l->u.string) == 0) {
+	    *out = l->name;
+	    return 1;
+	}
+	l = l->next;
+    }
+    return 0;
+}
+
+static int
 name_convert(krb5_context context, const char *name, const char *realm, 
-	     const char *section)
+	     const char **out)
 {
     const krb5_config_binding *l;
-    l = krb5_config_get_list (context->cf,
+    l = krb5_config_get_list (context,
+			      NULL,
 			      "realms",
 			      realm,
 			      "v4_name_convert",
-			      section,
+			      "host",
 			      NULL);
-    if(l == NULL)
-	l = krb5_config_get_list (context->cf,
-				  "libdefaults",
-				  "v4_name_convert",
-				  section,
-				  NULL);
+    if(l && check_list(l, name, out))
+	return KRB5_NT_SRV_HST;
+    l = krb5_config_get_list (context,
+			      NULL,
+			      "libdefaults",
+			      "v4_name_convert",
+			      "host",
+			      NULL);
+    if(l && check_list(l, name, out))
+	return KRB5_NT_SRV_HST;
+    l = krb5_config_get_list (context,
+			      NULL,
+			      "realms",
+			      realm,
+			      "v4_name_convert",
+			      "plain",
+			      NULL);
+    if(l && check_list(l, name, out))
+	return KRB5_NT_UNKNOWN;
+    l = krb5_config_get_list (context,
+			      NULL,
+			      "libdefaults",
+			      "v4_name_convert",
+			      "host",
+			      NULL);
+    if(l && check_list(l, name, out))
+	return KRB5_NT_UNKNOWN;
     
-    while(l){
-	if (l->type != STRING)
-	    continue;
-	if(strcmp(name, l->u.string) == 0)
-	    return l->name;
-	l = l->next;
+    /* didn't find it in config file, try built-in list */
+    {
+	struct v4_name_convert *q;
+	for(q = default_v4_name_convert; q->from; q++) {
+	    if(strcmp(name, q->to) == 0) {
+		*out = q->from;
+		return KRB5_NT_SRV_HST;
+	    }
+	}
     }
-    return NULL;
+    return -1;
 }
 
+/*
+ * convert the v5 principal in `principal' into a v4 corresponding one
+ * in `name, instance, realm'
+ * this is limited interface since there's no length given for these
+ * three parameters.  They have to be 40 bytes each (ANAME_SZ).
+ */
+
 krb5_error_code
 krb5_524_conv_principal(krb5_context context,
 			const krb5_principal principal,
@@ -554,9 +871,10 @@ krb5_524_conv_principal(krb5_context context,
 			char *instance,
 			char *realm)
 {
-    char *n, *i, *r;
+    const char *n, *i, *r;
     char tmpinst[40];
     int type = princ_type(principal);
+    const int aname_sz = 40;
 
     r = principal->realm;
 
@@ -574,41 +892,36 @@ krb5_524_conv_principal(krb5_context context,
     }
 
     {
-	char *tmp = name_convert(context, n, r, "host");
-	if(tmp){
-	    type = KRB5_NT_SRV_HST;
+	const char *tmp;
+	int t = name_convert(context, n, r, &tmp);
+	if(t >= 0) {
+	    type = t;
 	    n = tmp;
-	}else{
-	    tmp = name_convert(context, n, r, "plain");
-	    if(tmp){
-		type = KRB5_NT_UNKNOWN;
-		n = tmp;
-	    }
 	}
     }
 
     if(type == KRB5_NT_SRV_HST){
 	char *p;
-	strncpy(tmpinst, i, sizeof(tmpinst));
-	tmpinst[sizeof(tmpinst) - 1] = 0;
+
+	strlcpy (tmpinst, i, sizeof(tmpinst));
 	p = strchr(tmpinst, '.');
-	if(p) *p = 0;
+	if(p)
+	    *p = 0;
 	i = tmpinst;
     }
     
-    if(strlen(r) >= 40)
+    if (strlcpy (name, n, aname_sz) >= aname_sz)
 	return KRB5_PARSE_MALFORMED;
-    if(strlen(n) >= 40)
+    if (strlcpy (instance, i, aname_sz) >= aname_sz)
 	return KRB5_PARSE_MALFORMED;
-    if(strlen(i) >= 40)
+    if (strlcpy (realm, r, aname_sz) >= aname_sz)
 	return KRB5_PARSE_MALFORMED;
-    strcpy(realm, r);
-    strcpy(name, n);
-    strcpy(instance, i);
     return 0;
 }
 
-
+/*
+ * Create a principal in `ret_princ' for the service `sname' running
+ * on host `hostname'.  */
 			
 krb5_error_code
 krb5_sname_to_principal (krb5_context context,
@@ -618,36 +931,32 @@ krb5_sname_to_principal (krb5_context context,
 			 krb5_principal *ret_princ)
 {
     krb5_error_code ret;
-    char localhost[128];
+    char localhost[MAXHOSTNAMELEN];
     char **realms, *host = NULL;
-    
+	
     if(type != KRB5_NT_SRV_HST && type != KRB5_NT_UNKNOWN)
 	return KRB5_SNAME_UNSUPP_NAMETYPE;
-    if(hostname == NULL){
+    if(hostname == NULL) {
 	gethostname(localhost, sizeof(localhost));
 	hostname = localhost;
     }
     if(sname == NULL)
 	sname = "host";
-    if(type == KRB5_NT_SRV_HST){
-	struct hostent *hp;
-	hp = gethostbyname(hostname);
-	if(hp != NULL)
-	    hostname = hp->h_name;
-    }
-    ret = krb5_get_host_realm(context, hostname, &realms);
-    if(ret)
-	return ret;
-    if(type == KRB5_NT_SRV_HST){
-	host = strdup(hostname);
-	if(host == NULL){
-	    krb5_free_host_realm(context, realms);
-	    return ENOMEM;
-	}
+    if(type == KRB5_NT_SRV_HST) {
+	ret = krb5_expand_hostname_realms (context, hostname,
+					   &host, &realms);
+	if (ret)
+	    return ret;
 	strlwr(host);
 	hostname = host;
+    } else {
+	ret = krb5_get_host_realm(context, hostname, &realms);
+	if(ret)
+	    return ret;
     }
-    ret = krb5_make_principal(context, ret_princ, realms[0], sname, hostname, NULL);
+
+    ret = krb5_make_principal(context, ret_princ, realms[0], sname,
+			      hostname, NULL);
     if(host)
 	free(host);
     krb5_free_host_realm(context, realms);
